﻿/// File: PdfItfc.cs 
/// Date: 02/24/2016
/// Author: Phil Allison (pallison@stoneguardian.com)
/// Purpose: These are general support functions wrapped within a class in the base namespace for this library
/// 
/// copyright (c) 2001-2016, Stone Guardian Technologies, LLC. Under the MIT License http://www.opensource.org/licenses/mit-license.php  
/// Feel free to reuse and modify this code as long as credit is given to the original author.
/// Modified code must be noted in this section.
///  This code is based on original source from David Heffernan via http://stackoverflow.com/questions/21251895/using-ghostscript-9-10-in-windows-with-unicode-characters-in-parameters/21254325#21254325
///  - his original code was for UTF-8, which is included here but not implemented currently. 
///  
/// --- See the Readme.txt for any runtime issues. --- 
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Runtime.InteropServices;
using System.IO;
using System.Reflection;

namespace PDFUtil
{
	/// <summary>
	/// PDF Helper class, using ghostscript to compress PDF files. 
	/// </summary>
	public class PdfItfc
	{
		// this program could be run on either a 32 bit or 64 bit machine. 
		// 
		private static object _resourceLock = new object(); // we need the resource lock in order to make sure the Ghostscript DLL is used single threaded.
		private static bool _is32Bit = false;               // this flag is used to tell us if we are in 32 or 64 bit mode.
		public const int GS_ARG_ENCODING_LOCAL = 0;
		public const int GS_ARG_ENCODING_UTF8 = 1;

		#region dll import apis
		// This one module can handle both the 64 bit and 32 bit versions of operating systems. It just takes a little work within 
		// the code to test for which version and then call the appropriate functions

		//64 bit entry points
		[DllImport("gsdll64.dll", EntryPoint = "gsapi_new_instance")]
		private static extern int gsapi_new_instance(out IntPtr inst, IntPtr handle);

		[DllImport("gsdll64.dll", EntryPoint = "gsapi_set_arg_encoding")]
		private static extern int gsapi_set_arg_encoding(IntPtr inst, int encoding);

		[DllImport("gsdll64.dll", EntryPoint = "gsapi_init_with_args")]
		private static extern int gsapi_init_with_args(IntPtr inst, int argc, string[] argv); //if utf8 is needed, change argv to IntPtr[]

		[DllImport("gsdll64.dll", EntryPoint = "gsapi_init_with_args")]
		private static extern int gsapi_init_with_args8(IntPtr inst, int argc, IntPtr[] argv); //if utf8 is needed, change argv to IntPtr[]

		[DllImport("gsdll64.dll", EntryPoint = "gsapi_exit")]
		private static extern int gsapi_exit(IntPtr inst);

		[DllImport("gsdll64.dll")]
		private static extern void gsapi_delete_instance(IntPtr inst);

		//32 bit entry points
		[DllImport("gsdll32.dll", EntryPoint = "gsapi_new_instance")]
		private static extern int gsapi_new_instance32(out IntPtr inst, IntPtr handle);

		[DllImport("gsdll32.dll", EntryPoint = "gsapi_set_arg_encoding")]
		private static extern int gsapi_set_arg_encoding32(IntPtr inst, int encoding);

		[DllImport("gsdll32.dll", EntryPoint = "gsapi_init_with_args")]
		private static extern int gsapi_init_with_args32(IntPtr inst, int argc, string[] argv); //if utf8 is needed, change argv to IntPtr[]

		[DllImport("gsdll32.dll", EntryPoint = "gsapi_init_with_args")]
		private static extern int gsapi_init_with_args832(IntPtr inst, int argc, IntPtr[] argv); //if utf8 is needed, change argv to IntPtr[]

		[DllImport("gsdll32.dll", EntryPoint = "gsapi_exit")]
		private static extern int gsapi_exit32(IntPtr inst);

		[DllImport("gsdll32.dll", EntryPoint = "gsapi_delete_instance")]
		private static extern void gsapi_delete_instance32(IntPtr inst);

		#endregion

		/// <summary>
		/// Compress a PDF file using pre-optimized settings.
		/// </summary>
		/// <param name="fullFilePath">Full path and file name for the PDF.</param>
		/// <param name="destFile">Optional desitnation path and file name for the resulting compressed file.</param>
		/// <returns></returns>
		public static bool CompressFile(string fullFilePath, string destFile = "", bool copyOrigIfDestLarger = true)
		{
			if (destFile == "")
				destFile = fullFilePath;

			//we need to establish is the Ghostscript dlls folder are in the search path.
			string libpath = Program.PhysicalPath;
			//if(libpath.Contains("bin") == false) //this is more for use with Web based services.
			//	libpath += "bin\\";

			libpath += "lib";

			string path = Environment.GetEnvironmentVariable("Path", EnvironmentVariableTarget.Process);
			if (!path.Contains(libpath))
			{
				//add the lib folder to the path for this process
				Environment.SetEnvironmentVariable("Path", path + ";" + libpath, EnvironmentVariableTarget.Process);
			}

			_is32Bit = (IntPtr.Size == 4); //if IntPtr size is 4, we are on a 32 bit OS, otherwise we are on a 64 bit.

			bool ret = true;

			string infile = fullFilePath;
			string outfile = fullFilePath + "_";	//assign a temporary file for the outfile, 
													// this helps if there are any issues with the destination file or location.
			string[] args = new string[11];

			//the below parameters are all according to research into PDF optimization that I have done. more explanations soon
			// for full documentation on API arguments, see http://www.ghostscript.com/doc/9.18/Use.htm#Help_command, #10 command line options
			args[0] = "-Q"; //this argument is ignored by the API and can be anything 
			args[1] = "-P-";
			args[2] = "-dSAFER";
			args[3] = "-dNOPAUSE";
			args[4] = "-dBATCH";
			args[5] = "-sDEVICE=pdfwrite";
			args[6] = "-sCompression=lzw";
			args[7] = "-sOutputFile=" + outfile;
			args[8] = "-c";
			args[9] = ".setpdfwrite";
			args[10] = "-f" + infile;

			runApi(args);

			try
			{
				if (File.Exists(outfile))
				{
					if (copyOrigIfDestLarger)
					{
						//if original file is smaller, then do not use the new file. No compression happened.
						FileInfo fis = new FileInfo(infile);
						FileInfo fio = new FileInfo(outfile);

						if (fis.Length < fio.Length)
						{
							File.Copy(infile, outfile, true);
						}
					}

					File.Copy(outfile, destFile, true);
					File.Delete(outfile);
				}
				else
					ret = false;
			}
			catch
			{
				ret = false;
			}
			return ret;
		}

		#region private functions
		/// <summary>
		/// call the Ghostscript APIs with the given arguments
		/// </summary>
		/// <param name="argv">array of arguments to be passed into the APIs</param>
		private static void runApi(string[] argv)
		{
			IntPtr inst;
			int retcode = 0;
			lock (_resourceLock)
			{
				if (_is32Bit)
				{
					retcode = gsapi_new_instance32(out inst, IntPtr.Zero);
					try
					{
						retcode = gsapi_set_arg_encoding32(inst, GS_ARG_ENCODING_UTF8);
						retcode = gsapi_init_with_args32(inst, argv.Length, argv);
						retcode = gsapi_exit32(inst);
					}
					finally
					{
						gsapi_delete_instance32(inst);
					}
				}
				else //the system is 64-bit
				{
					retcode = gsapi_new_instance(out inst, IntPtr.Zero);
					try
					{
						retcode = gsapi_set_arg_encoding(inst, GS_ARG_ENCODING_UTF8);
						retcode = gsapi_init_with_args(inst, argv.Length, argv);
						retcode = gsapi_exit(inst);
					}
					finally
					{
						gsapi_delete_instance(inst);
					}
				}
			}
		}

		private static void runUtf8(string[] argv)
		{
			IntPtr inst;
			int retcode = 0;
			lock (_resourceLock)
			{
				retcode = gsapi_new_instance(out inst, IntPtr.Zero);
				try
				{
					IntPtr[] utf8argv = new IntPtr[argv.Length];
					for (int i = 0; i < utf8argv.Length; i++)
						utf8argv[i] = NativeUtf8FromString(argv[i]);
					try
					{
						retcode = gsapi_set_arg_encoding(inst, GS_ARG_ENCODING_UTF8);
						retcode = gsapi_init_with_args8(inst, utf8argv.Length, utf8argv);
						retcode = gsapi_exit(inst);
					}
					finally
					{
						gsapi_delete_instance(inst);
					}
				}
				finally
				{
					gsapi_delete_instance(inst);
				}
			}
		}
		/// <summary>
		/// Make a UTF8 pointer from a string. 
		/// Written by Hans Passant
		/// </summary>
		/// <param name="managedString"></param>
		/// <returns></returns>
		private static IntPtr NativeUtf8FromString(string managedString)
		{
			int len = Encoding.UTF8.GetByteCount(managedString);
			byte[] buffer = new byte[len + 1]; // null-terminator allocated
			Encoding.UTF8.GetBytes(managedString, 0, managedString.Length, buffer, 0);
			IntPtr nativeUtf8 = Marshal.AllocHGlobal(buffer.Length);
			Marshal.Copy(buffer, 0, nativeUtf8, buffer.Length);
			return nativeUtf8;
		}
		#endregion
	}
}
