﻿/// File:	frmPdfUtil.cs
/// Date:	02/24/2016
/// Author: Phil Allison (pallison@stoneguardian.com)
/// Purpose: basic form for adding files to a queue for PDF compression
/// 
/// copyright (c) 2001-2016, Stone Guardian Technologies, LLC. Under the MIT License http://www.opensource.org/licenses/mit-license.php  
/// Feel free to reuse and modify this code as long as credit is given to the original author.
/// Modified code must be noted in this section.
/// 
/// --- See the Readme.txt for any runtime issues. --- 
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using System.IO;
using System.Diagnostics;

namespace PDFUtil
{
	public partial class frmPdfUtil : Form
	{
		public frmPdfUtil()
		{
			InitializeComponent();
		}

		private void frmPdfUtil_Load(object sender, EventArgs e)
		{
			dgPdfs.AutoGenerateColumns = false;
			dgPdfs.AutoSizeRowsMode = DataGridViewAutoSizeRowsMode.DisplayedCells;

			//create an output folder if it does not exist
			string userDocs = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
			string output = userDocs + "\\PdfOutput";

			if (!Directory.Exists(output))
			{
				Directory.CreateDirectory(output);
			}
			lbOutputPath.Text = output;

			btCompress.Enabled = (SharedData.PdfFiles.Count > 0);
			setMessage("Add files or folders to the list.");
		}

		//Almost all of the menu items are just pass-through to the button click functions
		#region menu item actions

		private void addFileToolStripMenuItem_Click(object sender, EventArgs e)
		{
			btAddFile_Click(sender, e);
		}

		private void exitToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Application.Exit();
		}

		private void addFolderToolStripMenuItem_Click(object sender, EventArgs e)
		{
			btAddFolder_Click(sender, e);
		}

		private void removeSelectedToolStripMenuItem_Click(object sender, EventArgs e)
		{
			btRemoveSelected_Click(sender, e);
		}

		private void clearListToolStripMenuItem_Click(object sender, EventArgs e)
		{
			btClear_Click(sender, e);
		}

		private void openLocationToolStripMenuItem_Click(object sender, EventArgs e)
		{
			btOpenFolder_Click(sender, e);
		}

		private void changeLocationToolStripMenuItem_Click(object sender, EventArgs e)
		{
			btChangeDest_Click(sender, e);
		}

		private void compressFilesToolStripMenuItem_Click(object sender, EventArgs e)
		{
			btCompress_Click(sender, e);
		}

		private void onlineHelpToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Process.Start("http://stoneguardian.com/sgs/pdfutil.html#help");
		}

		private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
		{
			frmAbout fa = new frmAbout();
			fa.ShowDialog(this);
		}
		#endregion

		#region button actions

		/// <summary>
		/// Add a file to the list of files to compress
		/// </summary>
		private void btAddFile_Click(object sender, EventArgs e)
		{
			OpenFileDialog dialog = new OpenFileDialog();
			dialog.Multiselect = true;
			dialog.Filter = "PDF Files(.pdf)|*.pdf";
			if (dialog.ShowDialog() != DialogResult.OK)
				return;

			string[] fnames = dialog.FileNames;
			foreach (var fname in fnames)
			{ 
				if (!File.Exists(fname))
				{
					MessageBox.Show("The file you entered does not exist.", "File not found", MessageBoxButtons.OK);
					return;
				}
				addPdf(fname);
			}
			refreshList();
			setMessage("Add files or folders to the list, or press 'Compress Files' to compress the files in the list.");
		}

		/// <summary>
		/// add files from a specific folder to the list of files to compress
		/// </summary>
		private void btAddFolder_Click(object sender, EventArgs e)
		{
			FolderBrowserDialog dialog = new FolderBrowserDialog();
			if (dialog.ShowDialog() != DialogResult.OK)
				return;

			string[] files = Directory.GetFiles(dialog.SelectedPath, "*.pdf");

			foreach (var fname in files)
			{
				addPdf(fname);
			}
			refreshList();

		}

		/// <summary>
		/// remove selected file from the list to be compressed
		/// </summary>
		private void btRemoveSelected_Click(object sender, EventArgs e)
		{
			if (dgPdfs.CurrentRow == null)
				return;

			string srcDir = dgPdfs.CurrentRow.Cells[1].Value.ToString();

			PdfFile pdf = SharedData.PdfFiles.Find(x => x.SourceFolder == srcDir);

			if (pdf != null)
				SharedData.PdfFiles.Remove(pdf);

			refreshList();
		}

		/// <summary>
		/// Clear all files from the list
		/// </summary>
		private void btClear_Click(object sender, EventArgs e)
		{
			SharedData.PdfFiles.Clear();
			refreshList();
		}

		/// <summary>
		/// Open the current output folder
		/// </summary>
		private void btOpenFolder_Click(object sender, EventArgs e)
		{
			Process.Start(lbOutputPath.Text);
		}

		/// <summary>
		/// change the output folder for the compressed files
		/// </summary>
		private void btChangeDest_Click(object sender, EventArgs e)
		{
			FolderBrowserDialog dialog = new FolderBrowserDialog();
			if (dialog.ShowDialog() != DialogResult.OK)
				return;

			lbOutputPath.Text = dialog.SelectedPath;
		}

		/// <summary>
		/// run the compression
		/// </summary>
		private void btCompress_Click(object sender, EventArgs e)
		{
			SharedData.Compressing = true;
			setMessage("Compressing");
			tmAnimateMessage.Enabled = true;
			Application.UseWaitCursor = true;

			// start the actual compression on a thread so that updates can be made 
			// to the user interface as files are being compressed.
			Thread thread = new Thread(new ThreadStart(compressFiles));
			thread.Start();
		}
		#endregion

		#region logic functions

		/// <summary>
		/// Add given Pdf file to list of files to be compressed. 
		/// </summary>
		/// <param name="fname"></param>
		private void addPdf(string fname)
		{
			FileInfo fi = new FileInfo(fname);

			PdfFile pf = new PdfFile()
			{
				SourceFolder = fname,
				SourceFile = fi.Name,
				SourceSize = fi.Length,
				SourceSizeDisp = bytesToKB(fi.Length)
			};
			SharedData.PdfFiles.Add(pf);
		}

		/// <summary>
		/// run through list of all files and compress each one
		/// </summary>
		private void compressFiles()
		{
			FileInfo fi;
			SharedData.PdfFiles.ForEach(x =>
			{
				x.Converted = "●●●";

				x.DestFile = lbOutputPath.Text + "\\" + x.SourceFile;
				if (PdfItfc.CompressFile(x.SourceFolder, x.DestFile))
				{
					x.Converted = "Complete";
					fi = new FileInfo(x.DestFile);
					x.DestSize = fi.Length;
					x.DestSizeDisp = bytesToKB(x.DestSize);
					if (x.DestSize == x.SourceSize)
					{
						x.Ratio = SharedData.ZERO_RATIO;
					}
					else
					{
						double ratio = x.DestSize;
						ratio = ratio / x.SourceSize;
						x.Ratio = (100 - (ratio * 100)).ToString("N0") + "%";
					}
				}
				else
					x.Converted = "Failed";
			});

			SharedData.Compressing = false;
		}

		#endregion

		#region helper functions

		/// <summary>
		/// Timer is being used to check on current actions and update the user interface
		/// </summary>
		private void tmAnimateMessage_Tick(object sender, EventArgs e)
		{
			animateMessage();
			refreshList(true);

			if (SharedData.Compressing == false)
			{
				tmAnimateMessage.Enabled = false;
				Application.UseWaitCursor = false;
				setMessage("Compression Complete.");

				PdfFile pdf = SharedData.PdfFiles.Find(x => x.Ratio == SharedData.ZERO_RATIO);
				if (pdf != null)
				{
					setMessage("Compression Complete. Files marked by Red were not able to be compressed any further.");
				}
			}
		}

		/// <summary>
		/// refresh list of files compressed or to be compressed
		/// </summary>
		/// <param name="forceDisabled"></param>
		private void refreshList(bool forceDisabled = false)
		{
			dgPdfs.DataSource = null;
			dgPdfs.DataSource = SharedData.PdfFiles;
			Application.DoEvents();

			if (forceDisabled)
				btCompress.Enabled = false;
			else
			{
				btCompress.Enabled = (SharedData.PdfFiles.Count > 0);
				if (SharedData.PdfFiles.Count > 0)
					setMessage("Add files or folders to the list, or press 'Compress Files' to compress the files in the list.");
			}

			foreach (DataGridViewRow dgr in dgPdfs.Rows)
			{
				PdfFile pdf = (PdfFile)dgr.DataBoundItem;
				if (pdf.Ratio != null && pdf.Ratio == SharedData.ZERO_RATIO)
				{
					dgr.DefaultCellStyle.BackColor = Color.Red;
					dgr.DefaultCellStyle.ForeColor = Color.White;
				}
			}
		}

		/// <summary>
		/// simple text animation that adds ● to the end of the current message
		/// </summary>
		private void animateMessage()
		{
			string msg = lbMessage.Text;
			if (msg.Contains("●●●"))
				msg = msg.Trim('●');
			else
				msg += "●";
			setMessage(msg);
		}

		/// <summary>
		/// Set the current displayed message on the ui
		/// </summary>
		/// <param name="msg"></param>
		private void setMessage(string msg)
		{
			lbMessage.Text = msg;
			Application.DoEvents();
		}

		/// <summary>
		/// Convert filesize from bytes to kilobytes for readability
		/// </summary>
		/// <param name="len"></param>
		/// <returns></returns>
		private string bytesToKB(long len)
		{
			string ret = "";
			if (len < 1024)
				return "1KB";

			long kb = len / 1024;
			ret = kb.ToString("N0") + "KB";
			return ret;
		}

		#endregion
	}

	/// <summary>
	/// SharedData - a 'global' class to holh singleton instances of data across threads.
	/// </summary>
	public class SharedData
	{
		public const string ZERO_RATIO = "0%";

		// list of files that will be or have been compressed.
		// this is the source for the list on the ui, making the ui data driven
		public static List<PdfFile> PdfFiles = new List<PdfFile>();

		// flag to indicate across threads that the list of files is being processed. 
		public static bool Compressing = false;
	}

	/// <summary>
	/// PdfFile - a class to hold data related to each PDF file to be or that has been compressed.
	/// </summary>
	public class PdfFile
	{
		public string SourceFile { get; set; }
		public string SourceFolder { get; set; }
		public long SourceSize { get; set; }
		public string SourceSizeDisp { get; set; }
		public string DestFile { get; set; }
		public long DestSize { get; set; }
		public string DestSizeDisp { get; set; }
		public string Converted { get; set; }
		public string Ratio { get; set; }
	}
}
